"""Console script for glob_linters."""
import argparse
import logging
import os
import sys

from glob_linters.utils import io, settings

logger = logging.getLogger()


def _parse_args(args: list[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-d",
        "--target-dir",
        dest="target_dir",
        default=settings.Configs.target_dir,
        type=str,
        help="Target directory containing files to be linted",
    )
    parser.add_argument(
        "-s",
        "--file-suffix",
        dest="target_suffix",
        default=" ".join(settings.Configs.target_suffix),
        nargs="*",
        type=str,
        help="Target file extensions, like '.cpp', '.py'",
    )
    parser.add_argument(
        "-g",
        "--enable-debug",
        dest="debug",
        action="store_true",
        help="Enable debug mode, output debugging information",
    )
    parser.add_argument(
        "-c",
        "--configs",
        dest="configs",
        default=None,
        nargs="+",
        help="Set configuration, use `key=value` format and"
        "separate multiple pairs by comma or space",
    )
    parser.add_argument(
        "-C",
        "--config-file",
        dest="config_file",
        default=settings.DEFAULT_CONFIG_FILE_PATH,
        help="glob_linters configuration file (glob-linter.ini) path",
    )
    return parser.parse_args(args)


def _parse_config() -> None:
    args = _parse_args(sys.argv[1:])

    if os.path.exists(settings.DEFAULT_CONFIG_FILE_PATH):
        settings.Configs.has_read_config_file = True
        settings.parse_config_file(settings.DEFAULT_CONFIG_FILE_PATH)
    elif len(sys.argv) > 1:
        settings.parse_args(args)


def lint(targets: dict[str, list[str]]) -> None:
    """Linting process

    Parameters
    ----------
    targets : dict[str, list[str]]
        Files as a list to be linted for each file suffix
    """
    for ext, filenames in targets.items():
        for linter_name in settings.Configs.linters_enabled[ext]:
            for filename in filenames:
                logger.info("-" * 120)
                settings.Configs.return_code |= getattr(
                    settings.Configs, linter_name
                ).lint(filename)


def _set_logger() -> None:
    # logger = logging.getLogger()
    if settings.Configs.debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    stream = logging.StreamHandler()
    stream.setFormatter(
        logging.Formatter("%(asctime)s - [%(levelname)s] : %(message)s")
    )
    logger.addHandler(stream)


def main() -> int:
    """Console script for glob_linters."""
    _parse_config()
    _set_logger()

    logger.info("=" * 120)
    settings.load_linter_configs()
    logger.info("-" * 120)
    settings.install_mypy_package_requirements()

    logger.info("=" * 120)
    if settings.Configs.has_read_config_file:
        logger.info("Configuration file found, used that")
    else:
        logger.info("Configuration file not found, used defaults or command arguments")
    io.print_configs()

    # Scan files
    logger.info("Starting directory scan: %s", settings.Configs.target_dir)
    logger.info("Target suffix:")
    for suffix in settings.Configs.target_suffix:
        logger.info("\t%s", suffix)
    target_files = io.scan(settings.Configs.target_dir, settings.Configs.target_suffix)
    logger.info("Target file list:")
    for lang, files in target_files.items():
        for file in files:
            logger.info("\t%s - %s", lang, file)

    # Linting
    logger.info("=" * 120)
    logger.info("Lint starting...")
    lint(target_files)

    return settings.Configs.return_code


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



"""IO functions"""

import logging
import os
from pathlib import Path

from glob_linters.utils import settings

logger = logging.getLogger(__name__)


def print_configs() -> None:
    """Show configuration"""
    attr_names_sorted_logically = [
        "has_read_config_file",
        "debug",
        "linters_enabled",
        "target_dir",
        "target_suffix",
    ]
    logger.debug("Configuration set:")
    for attr in attr_names_sorted_logically:
        if attr == "linters_enabled":
            for ext in settings.Configs.target_suffix:
                for linter_name in settings.Configs.linters_enabled[ext]:
                    logger.debug(
                        "\t%s - %s executable: %s, use configuration file: %s",
                        ext,
                        linter_name,
                        getattr(settings.Configs, linter_name).executable,
                        getattr(settings.Configs, linter_name).use_config_file,
                    )
        else:
            logger.debug("\t%s: %s", attr, getattr(settings.Configs, attr))


def scan(target_dir: str, suffix: list[str]) -> dict[str, list[str]]:
    """Scan directories to obtain target files

    Parameters
    ----------
    target_dirs : list[str]
        Directories to be scanned
    suffix : list[str]
        Expected file suffix

    Returns
    -------
    dict[str, list[str]]
        Absolute paths of target files
    """
    target_files: dict[str, list[str]] = {s: [] for s in suffix}
    target_dir = os.path.abspath(target_dir)
    logger.debug("Scanning directory: %s", target_dir)
    for dirpath, _, filenames in os.walk(target_dir):
        for filename in filenames:
            f_path = Path(os.path.join(dirpath, filename))
            logger.debug("Found file: %s", f_path)
            if f_path.suffix in suffix:
                logger.debug("Found qualified file: %s", f_path)
                target_files[f_path.suffix].append(str(f_path))
    return target_files



"""Linters"""
# pylint: disable=subprocess-run-check

import logging
import subprocess

logger = logging.getLogger(__name__)


class Linter:
    """Linter parent class"""

    def __init__(self, executable: str) -> None:
        self.executable = executable
        self.options: list[str] = []
        self.cmd_result: subprocess.CompletedProcess
        self.stdout: list[str]
        self.stderr: list[str]
        self.use_config_file: bool = False

    def lint(self, filename: str) -> int:
        """General linting method

        Parameters
        ----------
        filename : str
            File path to be linted

        Returns
        -------
        int
            Return code of the linter program
        """
        logger.info("Linting with [%s] on file %s", self.executable, filename)
        cmd = [self.executable] + self.options + [filename]
        logger.debug("Linting command: %s", " ".join(cmd))
        self.cmd_result = subprocess.run(cmd, capture_output=True)
        self.stdout = self.cmd_result.stdout.decode().strip().split("\n")
        self.stderr = self.cmd_result.stderr.decode().strip().split("\n")

        self.process_output()
        return self.cmd_result.returncode

    def process_output(self) -> None:
        """Process output, since some linters print errors to stdout"""
        logger.debug("Linter stdout:")
        for out in self.stdout:
            logger.debug("\t%s", out)

        if self.cmd_result.returncode != 0:
            logger.error("Found errors:")
            for err in self.stderr:
                logger.error("\t%s", err)
        else:
            logger.info("Check passed.")


# Linters for c/c++
class ClangFormat(Linter):
    """`clang-format` linter"""

    def __init__(self, executable: str) -> None:
        super().__init__(executable)
        self.options = ["--dry-run", "--Werror"]


class Cpplint(Linter):
    """`cpplint` linter"""


# Linters for Python
class Pylint(Linter):
    """`pylint` linter"""

    def __init__(self, executable: str) -> None:
        super().__init__(executable)
        self.options = ["--output-format=parseable"]

    def process_output(self) -> None:
        if self.cmd_result.returncode != 0:
            logger.error("Found errors:")
            for err in self.stdout:
                logger.error("\t%s", err)
        else:
            logger.info("Check passed.")


class Flake8(Linter):
    """`flake8` linter"""

    def process_output(self) -> None:
        if self.cmd_result.returncode != 0:
            logger.error("Found errors:")
            for err in self.stdout:
                logger.error("\t%s", err)
        else:
            logger.info("Check passed.")


class Black(Linter):
    """`black` linter"""

    def __init__(self, executable: str) -> None:
        super().__init__(executable)
        self.options = ["--check", "--diff"]

    def process_output(self) -> None:
        if self.cmd_result.returncode != 0:
            logger.error("Found errors:")
            for err in self.stdout:
                logger.error("\t%s", err)
        else:
            logger.info("Check passed.")


class Isort(Linter):
    """`isort` linter"""

    def __init__(self, executable: str) -> None:
        super().__init__(executable)
        self.options = ["--check-only", "--diff"]

    def process_output(self) -> None:
        if self.cmd_result.returncode != 0:
            logger.error("Found errors:")
            for err in self.stdout:
                logger.error("\t%s", err)
        else:
            logger.info("Check passed.")


class Mypy(Linter):
    """`mypy` linter"""

    def __init__(self, executable: str) -> None:
        super().__init__(executable)
        self.options = ["--pretty", "--show-error-context", "--show-error-codes"]

    def process_output(self) -> None:
        if self.cmd_result.returncode != 0:
            logger.error("Found errors:")
            for err in self.stdout:
                logger.error("\t%s", err)
        else:
            logger.info("Check passed.")



"""Configuration"""
# Configuration vairables and defaults
# pylint: disable=subprocess-run-check
import argparse
import configparser
import logging
import os
import re
import subprocess
import sys
from dataclasses import dataclass
from typing import Callable, ClassVar

from glob_linters import linters

# Default config file
DEFAULT_CONFIG_FILE_PATH: str = ".github/glob-linters.ini"
MYPY_PACKAGE_REQUIREMENTS_FILE_PATH: str = os.path.abspath(
    ".github/mypy_requirements.txt"
)

# User-defined linter config directory for GitHub action
USER_DEFIEND_LINTER_CONFIG_ROOT: str = os.path.abspath(".github/linter-configs")

# Supported file extension/suffix
SUPPORTED_FILE_SUFFIX: list[str] = [".cpp", ".py"]

logger = logging.getLogger(__name__)


@dataclass
class Configs:
    """Class to store configs/settings"""

    # Exit code record
    return_code: ClassVar[int] = 0

    # Indicator of reading file
    has_read_config_file: ClassVar[bool] = False

    # Lint targets settings
    target_dir: ClassVar[str] = "."
    target_suffix: ClassVar[list[str]] = SUPPORTED_FILE_SUFFIX

    # Running mode
    debug: ClassVar[bool] = False

    # Linter dict for running
    # Clang Linters
    cpplint: ClassVar[linters.Cpplint] = linters.Cpplint("cpplint")
    clang_format: ClassVar[linters.ClangFormat] = linters.ClangFormat("clang-format")
    # Python
    pylint: ClassVar[linters.Pylint] = linters.Pylint("pylint")
    flake8: ClassVar[linters.Flake8] = linters.Flake8("flake8")
    black: ClassVar[linters.Black] = linters.Black("black")
    isort: ClassVar[linters.Isort] = linters.Isort("isort")
    mypy: ClassVar[linters.Mypy] = linters.Mypy("mypy")

    # Linters needed for each file extension
    linters_enabled: ClassVar[dict[str, list[str]]] = {
        ".cpp": ["cpplint", "clang_format"],
        ".py": ["pylint", "flake8", "black", "isort", "mypy"],
    }

    # Available configs
    set_configs: ClassVar[dict[str, dict[str, Callable]]] = {
        "target": {
            "target_dir": lambda x: setattr(Configs, "target_dir", x),
            "target_suffix": lambda x: setattr(
                Configs, "target_suffix", re.split(r"[,\s]", x)
            ),
        },
        "executable": {
            # Clang
            "cpplint": lambda x: setattr(Configs.cpplint, "executable", x),
            "cpplint.options": lambda x: setattr(
                Configs.cpplint,
                "options",
                Configs.cpplint.options + re.split(r"[,\s]", x),
            ),
            "clang_format": lambda x: setattr(Configs.clang_format, "executable", x),
            "clang_format.options": lambda x: setattr(
                Configs.clang_format,
                "options",
                Configs.clang_format.options + re.split(r"[,\s]", x),
            ),
            # Python
            "pylint": lambda x: setattr(Configs.pylint, "executable", x),
            "pylint.options": lambda x: setattr(
                Configs.pylint,
                "options",
                Configs.pylint.options + re.split(r"[,\s]", x),
            ),
            "flake8": lambda x: setattr(Configs.flake8, "executable", x),
            "flake8.options": lambda x: setattr(
                Configs.flake8,
                "options",
                Configs.flake8.options + re.split(r"[,\s]", x),
            ),
            "black": lambda x: setattr(Configs.black, "executable", x),
            "black.options": lambda x: setattr(
                Configs.black, "options", Configs.black.options + re.split(r"[,\s]", x)
            ),
            "isort": lambda x: setattr(Configs.isort, "executable", x),
            "isort.options": lambda x: setattr(
                Configs.isort, "options", Configs.isort.options + re.split(r"[,\s]", x)
            ),
            "mypy": lambda x: setattr(Configs.mypy, "executable", x),
            "mypy.options": lambda x: setattr(
                Configs.mypy, "options", Configs.mypy.options + re.split(r"[,\s]", x)
            ),
        },
        "env": {
            "debug": lambda x: setattr(Configs, "debug", x),
            ".cpp.linters": lambda x: getattr(Configs, "linters_enabled").update(
                {".cpp": re.split(r",\s", x)}
            ),
            ".cpp.disable_linters": lambda x: getattr(
                Configs, "linters_enabled"
            ).update(
                {
                    ".cpp": [
                        value
                        for value in Configs.linters_enabled[".cpp"]
                        if value not in re.split(r"[,\s]", x)
                    ]
                }
            ),
            ".py.linters": lambda x: getattr(Configs, "linters_enabled").update(
                {".py": re.split(r"[,\s]", x)}
            ),
            ".py.disable_linters": lambda x: getattr(Configs, "linters_enabled").update(
                {
                    ".py": [
                        value
                        for value in Configs.linters_enabled[".py"]
                        if value not in re.split(r"[,\s]", x)
                    ]
                }
            ),
        },
        "_set-config-file": {
            "cpplint": lambda: os.symlink(
                os.path.join(USER_DEFIEND_LINTER_CONFIG_ROOT, "CPPLINT.cfg"),
                "CPPLINT.cfg",
            )
            if os.path.exists(
                os.path.join(USER_DEFIEND_LINTER_CONFIG_ROOT, "CPPLINT.cfg")
            )
            else -1,
            "clang_format": lambda: os.symlink(
                os.path.join(USER_DEFIEND_LINTER_CONFIG_ROOT, ".clang-format"),
                ".clang-format",
            )
            if os.path.exists(
                os.path.join(USER_DEFIEND_LINTER_CONFIG_ROOT, ".clang-format")
            )
            else -1,
            "pylint": lambda: Configs.pylint.options.extend(
                ["--rcfile", os.path.join(USER_DEFIEND_LINTER_CONFIG_ROOT, ".pylintrc")]
            )
            if os.path.exists(
                os.path.join(USER_DEFIEND_LINTER_CONFIG_ROOT, ".pylintrc")
            )
            else -1,
            "flake8": lambda: Configs.flake8.options.extend(
                ["--config", os.path.join(USER_DEFIEND_LINTER_CONFIG_ROOT, ".flake8")]
            )
            if os.path.exists(os.path.join(USER_DEFIEND_LINTER_CONFIG_ROOT, ".flake8"))
            else -1,
            "black": lambda: Configs.black.options.extend(
                ["--config", os.path.join(USER_DEFIEND_LINTER_CONFIG_ROOT, ".black")]
            )
            if os.path.exists(os.path.join(USER_DEFIEND_LINTER_CONFIG_ROOT, ".black"))
            else -1,
            "isort": lambda: Configs.isort.options.extend(
                [
                    "--settings-file",
                    os.path.join(USER_DEFIEND_LINTER_CONFIG_ROOT, ".isort.cfg"),
                ]
            )
            if os.path.exists(
                os.path.join(USER_DEFIEND_LINTER_CONFIG_ROOT, ".isort.cfg")
            )
            else -1,
            "mypy": lambda: Configs.mypy.options.extend(
                [
                    "--config-file",
                    os.path.join(USER_DEFIEND_LINTER_CONFIG_ROOT, ".mypy.ini"),
                ]
            )
            if os.path.exists(
                os.path.join(USER_DEFIEND_LINTER_CONFIG_ROOT, ".mypy.ini")
            )
            else -1,
        },
    }


def parse_config_file(config_file: str) -> None:
    """Parse configuration from configparser-like file

    Parameters
    ----------
    config_file : str
        Configuration file path

    Raises
    ------
    ValueError
        Invalid section/option found in config file
    """
    config_parser = configparser.ConfigParser()
    config_parser.read(config_file)
    for section in config_parser:
        if section == "DEFAULT":
            continue
        if section not in Configs.set_configs:
            raise ValueError(f"No such section supported: {section}")
        for option in config_parser[section]:
            if option not in config_parser[section]:
                raise ValueError(f"No such option supported: {option}")
            if option == "target_suffix":
                diff = set(re.split(r"[,\s]", config_parser[section][option])) - set(
                    SUPPORTED_FILE_SUFFIX
                )
                if len(diff) > 0:
                    raise ValueError(f"No supported file suffix: {diff}")
            Configs.set_configs[section][option](config_parser[section][option])


def parse_args(args: argparse.Namespace) -> None:
    """Parse command line arguments

    Parameters
    ----------
    args : argparse.Namespace
        Arguments returned by `argparse.parse_args`

    Raises
    ------
    ValueError
        Invalid linter name found
    """
    if os.path.exists(args.config_file):
        parse_config_file(args.config_file)
    else:
        Configs.debug = args.debug
        Configs.target_dir = args.target_dir
        Configs.target_suffix = re.split(r"\s", args.target_suffix)
        if args.configs is not None:
            for key, value in map(lambda x: re.split("=", x), args.configs):
                if key in Configs.set_configs["executable"]:
                    if "options" in key:
                        Configs.set_configs["executable"][key](value)
                    else:
                        Configs.set_configs["executable"][key](value)
                elif key in Configs.set_configs["env"]:
                    Configs.set_configs["env"][key](value)
                else:
                    raise ValueError(f"No such linter supported: {key}")


def install_mypy_package_requirements() -> None:
    """Install additional packages for mypy linting"""
    if not os.path.exists(MYPY_PACKAGE_REQUIREMENTS_FILE_PATH):
        logger.info("Not found mypy_requirements.txt, skip package installation")
        return
    logger.info("Install packages for mypy checking...")
    cmd = ["pip", "install", "-r", MYPY_PACKAGE_REQUIREMENTS_FILE_PATH]
    logger.debug("Install command: %s", " ".join(cmd))
    cmd_result = subprocess.run(cmd, capture_output=True)

    logger.debug("Installation output:")
    for line in cmd_result.stdout.decode().strip().split("\n"):
        logger.debug("\t%s", line)

    if cmd_result.returncode != 0:
        logger.error("Package installaltion failed:")
        for line in cmd_result.stderr.decode().strip().split("\n"):
            logger.error("\t%s", line)
        sys.exit(1)


def load_linter_configs() -> None:
    """Load config file for each enabled linter"""
    for ext in Configs.target_suffix:
        for linter_name in Configs.linters_enabled[ext]:
            logger.debug("Setting linter config for [%s]", linter_name)
            # Check whether set config file
            # If set, return None, -1 otherwise
            return_value = Configs.set_configs["_set-config-file"][linter_name]()
            if return_value is None:
                getattr(Configs, linter_name).use_config_file = True



#!/usr/bin/env python

"""The setup script."""

from setuptools import find_packages, setup

with open("README.rst", encoding="utf-8") as readme_file:
    readme = readme_file.read()

with open("CHANGELOG.md", encoding="utf-8") as changelog_file:
    changelog = changelog_file.read()

setup_requirements = ["setuptools_scm"]

requirements = [
    "clang-format>=14.0.6",
    "cpplint>=1.6.1",
    "pylint>=2.15.2",
    "flake8>=5.0.4",
    "black>=22.8.0",
    "isort>=5.10.1",
    "mypy>=0.971",
]

test_requirements = [
    "pytest>=3",
]


def _local_scheme(version: str) -> str:  # pylint: disable=unused-argument
    """Adjust local version

    Parameters
    ----------
    version : str
        Version string

    Returns
    -------
    str
        Adjusted version
    """
    return ""


setup(
    author="Bowen Tan",
    author_email="bowentan78@gmail.com",
    setup_requires=setup_requirements,
    use_scm_version={"local_scheme": _local_scheme},
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="Integrated tool for multi-language linters for command line tool"
    "and GitHub action",
    entry_points={
        "console_scripts": [
            "glob-linters=glob_linters.cli:main",
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + changelog,
    long_description_content_type="text/x-rst",
    include_package_data=True,
    keywords="glob-linters",
    name="glob-linters",
    packages=find_packages(include=["glob_linters", "glob_linters.*"]),
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/bowentan/glob-linters",
    version="0.2.2",
    zip_safe=False,
)
